..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-scheduler:

.. list-table:: Description of scheduler configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``filter_function`` = ``None``
     - (String) String representation for an equation that will be used to filter hosts. Only used when the driver filter is set to be used by the Cinder scheduler.
   * - ``goodness_function`` = ``None``
     - (String) String representation for an equation that will be used to determine the goodness of a host. Only used when using the goodness weigher is set to be used by the Cinder scheduler.
   * - ``scheduler_default_filters`` = ``AvailabilityZoneFilter, CapacityFilter, CapabilitiesFilter``
     - (List) Which filter class names to use for filtering hosts when not specified in the request.
   * - ``scheduler_default_weighers`` = ``CapacityWeigher``
     - (List) Which weigher class names to use for weighing hosts.
   * - ``scheduler_driver`` = ``cinder.scheduler.filter_scheduler.FilterScheduler``
     - (String) Default scheduler driver to use
   * - ``scheduler_host_manager`` = ``cinder.scheduler.host_manager.HostManager``
     - (String) The scheduler host manager class to use
   * - ``scheduler_json_config_location`` =
     - (String) Absolute path to scheduler configuration JSON file.
   * - ``scheduler_manager`` = ``cinder.scheduler.manager.SchedulerManager``
     - (String) Full class name for the Manager for scheduler
   * - ``scheduler_max_attempts`` = ``3``
     - (Integer) Maximum number of attempts to schedule a volume
   * - ``scheduler_weight_handler`` = ``cinder.scheduler.weights.OrderedHostWeightHandler``
     - (String) Which handler to use for selecting the host/pool after weighing
