{
    "snapshot": {
        "name": "snap-001",
        "description": "Daily backup",
        "volume_id": "%(volume_id)s",
        "force": true,
        "metadata": {
            "key": "v3"
        }
    }
}
