..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-osbrick:

.. list-table:: Description of os-brick configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[privsep_osbrick]**
     -
   * - ``capabilities`` = ``[]``
     - (Unknown) List of Linux capabilities retained by the privsep daemon.
   * - ``group`` = ``None``
     - (String) Group that the privsep daemon should run as.
   * - ``helper_command`` = ``None``
     - (String) Command to invoke to start the privsep daemon if not using the "fork" method. If not specified, a default is generated using "sudo privsep-helper" and arguments designed to recreate the current configuration. This command must accept suitable --privsep_context and --privsep_sock_path arguments.
   * - ``user`` = ``None``
     - (String) User that the privsep daemon should run as.
