{
    "snapshot": {
        "created_at": "%(strtime)s",
        "description": "This is yet, another snapshot.",
        "id": "%(uuid)s",
        "metadata": {
            "key": "v3"
        },
        "name": "snap-002",
        "size": 10,
        "status": "creating",
        "updated_at": null,
        "volume_id": "%(uuid)s"
    }
}