..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-storage_gpfs:

.. list-table:: Description of GPFS storage configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``gpfs_images_dir`` = ``None``
     - (String) Specifies the path of the Image service repository in GPFS. Leave undefined if not storing images in GPFS.
   * - ``gpfs_images_share_mode`` = ``None``
     - (String) Specifies the type of image copy to be used. Set this when the Image service repository also uses GPFS so that image files can be transferred efficiently from the Image service to the Block Storage service. There are two valid values: "copy" specifies that a full copy of the image is made; "copy_on_write" specifies that copy-on-write optimization strategy is used and unmodified blocks of the image file are shared efficiently.
   * - ``gpfs_max_clone_depth`` = ``0``
     - (Integer) Specifies an upper limit on the number of indirections required to reach a specific block due to snapshots or clones. A lengthy chain of copy-on-write snapshots or clones can have a negative impact on performance, but improves space utilization. 0 indicates unlimited clone depth.
   * - ``gpfs_mount_point_base`` = ``None``
     - (String) Specifies the path of the GPFS directory where Block Storage volume and snapshot files are stored.
   * - ``gpfs_sparse_volumes`` = ``True``
     - (Boolean) Specifies that volumes are created as sparse files which initially consume no space. If set to False, the volume is created as a fully allocated file, in which case, creation may take a significantly longer time.
   * - ``gpfs_storage_pool`` = ``system``
     - (String) Specifies the storage pool that volumes are assigned to. By default, the system storage pool is used.
   * - ``nas_host`` =
     - (String) IP address or Hostname of NAS system.
   * - ``nas_login`` = ``admin``
     - (String) User name to connect to NAS system.
   * - ``nas_password`` =
     - (String) Password to connect to NAS system.
   * - ``nas_private_key`` =
     - (String) Filename of private key to use for SSH authentication.
   * - ``nas_ssh_port`` = ``22``
     - (Port number) SSH port to use to connect to NAS system.
