{
    "volume": {
        "attachments": [],
        "availability_zone": "az2",
        "bootable": "false",
        "created_at": "2014-07-18T00:12:54.000000",
        "description": "Volume imported from existingLV",
        "encrypted": "false",
        "id": "23cf872b-c781-4cd4-847d-5f2ec8cbd91c",
        "links": [
            {
                "href": "http://10.0.2.15:8776/v3/87c8522052ca4eed98bc672b4c1a3ddb/volumes/23cf872b-c781-4cd4-847d-5f2ec8cbd91c",
                "rel": "self"
            },
            {
                "href": "http://10.0.2.15:8776/87c8522052ca4eed98bc672b4c1a3ddb/volumes/23cf872b-c781-4cd4-847d-5f2ec8cbd91c",
                "rel": "bookmark"
            }
        ],
        "metadata": {
            "key1": "value1",
            "key2": "value2"
        },
        "name": "New Volume",
        "os-vol-tenant-attr:tenant_id": "87c8522052ca4eed98bc672b4c1a3ddb",
        "size": 0,
        "snapshot_id": "null",
        "source_volid": "null",
        "status": "creating",
        "user_id": "eae1472b5fc5496998a3d06550929e7e",
        "volume_type": "null"
    }
}