{
    "transfers": [
        {
            "id": "%(uuid)s",
            "links": [
                {
                    "href": "%(host)s/v3/%(uuid)s/os-volume-transfer/%(uuid)s",
                    "rel": "self"
                },
                {
                    "href": "%(host)s/%(uuid)s/os-volume-transfer/%(uuid)s",
                    "rel": "bookmark"
                }
            ],
            "name": "first volume",
            "volume_id": "%(uuid)s"
        }
    ]
}