..
    Warning: Do not edit this file. It is automatically generated from the
    software project's code and your changes will be overwritten.

    The tool to generate this file lives in openstack-doc-tools repository.

    Please make any changes needed in the code, then run the
    autogenerate-config-doc tool from the openstack-doc-tools repository, or
    ask for help on the documentation mailing list, IRC channel or meeting.

.. _cinder-storage:

.. list-table:: Description of storage configuration options
   :header-rows: 1
   :class: config-ref-table

   * - Configuration option = Default value
     - Description
   * - **[DEFAULT]**
     -
   * - ``allocated_capacity_weight_multiplier`` = ``-1.0``
     - (Floating point) Multiplier used for weighing allocated capacity. Positive numbers mean to stack vs spread.
   * - ``capacity_weight_multiplier`` = ``1.0``
     - (Floating point) Multiplier used for weighing free capacity. Negative numbers mean to stack vs spread.
   * - ``enabled_backends`` = ``None``
     - (List) A list of backend names to use. These backend names should be backed by a unique [CONFIG] group with its options
   * - ``target_helper`` = ``tgtadm``
     - (String) iSCSI target user-land tool to use. tgtadm is default, use lioadm for LIO iSCSI support, scstadmin for SCST target support, ietadm for iSCSI Enterprise Target, iscsictl for Chelsio iSCSI Target or fake for testing.
   * - ``iscsi_iotype`` = ``fileio``
     - (String) Sets the behavior of the iSCSI target to either perform blockio or fileio optionally, auto can be set and Cinder will autodetect type of backing device
   * - ``target_ip_address`` = ``$my_ip``
     - (String) The IP address that the iSCSI daemon is listening on
   * - ``target_port`` = ``3260``
     - (Port number) The port that the iSCSI daemon is listening on
   * - ``target_protocol`` = ``iscsi``
     - (String) Determines the iSCSI protocol for new iSCSI volumes, created with tgtadm or lioadm target helpers. In order to enable RDMA, this parameter should be set with the value "iser". The supported iSCSI protocol values are "iscsi" and "iser".
   * - ``iscsi_target_flags`` =
     - (String) Sets the target-specific flags for the iSCSI target. Only used for tgtadm to specify backing device flags using bsoflags option. The specified string is passed as is to the underlying tool.
   * - ``target_prefix`` = ``iqn.2010-10.org.openstack:``
     - (String) Prefix for iSCSI volumes
   * - ``iscsi_write_cache`` = ``on``
     - (String) Sets the behavior of the iSCSI target to either perform write-back(on) or write-through(off). This parameter is valid if target_helper is set to tgtadm.
   * - ``iser_helper`` = ``tgtadm``
     - (String) The name of the iSER target user-land tool to use
   * - ``iser_ip_address`` = ``$my_ip``
     - (String) The IP address that the iSER daemon is listening on
   * - ``iser_port`` = ``3260``
     - (Port number) The port that the iSER daemon is listening on
   * - ``iser_target_prefix`` = ``iqn.2010-10.org.openstack:``
     - (String) Prefix for iSER volumes
   * - ``migration_create_volume_timeout_secs`` = ``300``
     - (Integer) Timeout for creating the volume to migrate to when performing volume migration (seconds)
   * - ``num_iser_scan_tries`` = ``3``
     - (Integer) The maximum number of times to rescan iSER targetto find volume
   * - ``num_volume_device_scan_tries`` = ``3``
     - (Integer) The maximum number of times to rescan targets to find volume
   * - ``volume_backend_name`` = ``None``
     - (String) The backend name for a given driver implementation
   * - ``volume_clear`` = ``zero``
     - (String) Method used to wipe old volumes. This option can either be set in section [backend_defaults] (to apply to all enabled_backends) or directly in the backend specific section to apply for just this one backend.
   * - ``volume_clear_ionice`` = ``None``
     - (String) The flag to pass to ionice to alter the i/o priority of the process used to zero a volume after deletion, for example "-c3" for idle only priority. This option can either be set in section [backend_defaults] (to apply to all enabled_backends) or directly in the backend specific section to apply for just this one backend.
   * - ``volume_clear_size`` = ``0``
     - (Integer) Size in MiB to wipe at start of old volumes. 1024 MiBat max. 0 => all. This option can either be set in section [backend_defaults] (to apply to all enabled_backends) or directly in the backend specific section to apply for just this one backend.
   * - ``volume_copy_blkio_cgroup_name`` = ``cinder-volume-copy``
     - (String) The blkio cgroup name to be used to limit bandwidth of volume copy
   * - ``volume_copy_bps_limit`` = ``0``
     - (Integer) The upper limit of bandwidth of volume copy. 0 => unlimited
   * - ``volume_dd_blocksize`` = ``1M``
     - (String) The default block size used when copying/clearing volumes
   * - ``volume_driver`` = ``cinder.volume.drivers.lvm.LVMVolumeDriver``
     - (String) Driver to use for volume creation
   * - ``volume_manager`` = ``cinder.volume.manager.VolumeManager``
     - (String) Full class name for the Manager for volume
   * - ``volume_service_inithost_offload`` = ``False``
     - (Boolean) Offload pending volume delete during volume service startup
   * - ``volume_usage_audit_period`` = ``month``
     - (String) Time period for which to generate volume usages. The options are hour, day, month, or year.
   * - ``volumes_dir`` = ``$state_path/volumes``
     - (String) Volume configuration file storage directory
